#include "dictionary.h"

Dictionary::Dictionary(const vector<string>& dataset, const int& eps, const vector<bool>& passFlags, vector<double>& mmlMat, vector<double>& condMat, const string& logfile) {
    assert(dataset.size() > 0);
    assert(passFlags.size() >= 3);
    assert(mmlMat.size() == residueSize && condMat.size() == residueSize*residueSize);

    dataSet = dataset;
    mmlMatrix = mmlMat;
    condMatrix = condMat;
    epsln = eps;
    pertLegend = passFlags[0];
    seqCost = passFlags[1];
    alignPrint = passFlags[2];
    logF = logfile;

    double nullHypTotal = 0;
    double seqNull = 0;

    for(size_t p = 0; p < dataSet.size(); p++){
        seqNull = mmlEstimate(dataSet[p].substr(0, dataSet[p].size()), mmlMatrix) + encodeWTNits(dataSet[p].size());
        nullHypTotal += seqNull;
        cost.push_back(vector<double> (1, seqNull));
    }

    nullML = encodeWTNits(dataSet.size()) + nullHypTotal;
    concepts.push_back(-1);
    masks.push_back(true);

    nSize++;
    nSizeTrue++;
}

void Dictionary::_pop(){
	assert(nSize > 1 && concepts.size() == nSize && masks.size() == nSize);
    assert(cost.size() == dataSet.size() && cost[0].size() == nSize);

    string ccpt = dataSet[concepts[concepts.size()-1]];
    dictML -= mmlEstimate(ccpt, mmlMatrix) + encodeWTNits(ccpt.size());

    concepts.pop_back();
    masks.pop_back();

    for(int i = 0; i < cost.size(); i++)
        cost[i].pop_back();
    
    nSize--;
    nSizeTrue--;
}

int Dictionary::_randConcept(){
    assert(nSize > 1);
    int indx = rand() % (nSize-1) + 1;
    assert(masks[indx]);
    return indx;
}

void Dictionary::_suppressConcept(const int& indx){
    assert(indx > 0 && indx < nSize);
    assert(masks[indx]);
    
    masks[indx] = false;
    tempML = mmlEstimate(dataSet[concepts[indx]], mmlMatrix) + encodeWTNits(dataSet[concepts[indx]].size());
    dictML -= tempML;
    
    nSizeTrue--;
    
    if(pertLegend){
        string pertLeg = "Del : " + to_string(concepts[indx]) + "\n";
        file_append(logF, pertLeg);
        cout << pertLeg;
    }    
}

void Dictionary::_unsuppressConcept(const int& indx){
    assert(indx > 0 && indx < nSize);
    assert(!masks[indx]);

    masks[indx] = true;
    dictML += tempML;
    assert(abs(tempML - (mmlEstimate(dataSet[concepts[indx]], mmlMatrix) + encodeWTNits(dataSet[concepts[indx]].size())) < 1e-6));

    nSizeTrue++;
}

void Dictionary::_newConcept(int& indx){
    assert(dataSet.size() > 0);
    indx = rand() % dataSet.size();
    assert(indx >= 0 && indx < dataSet.size());
}

bool Dictionary::_add(int& seqIndx){
    assert(seqIndx >= 0 && seqIndx < dataSet.size());
    assert(cost.size() == dataSet.size());

    double seqML = 0;

    if(find(concepts.begin(), concepts.end(), seqIndx) != concepts.end()) return false;

    #pragma omp parallel for 
    for(size_t q = 0; q < dataSet.size(); q++){
        seqML = getAlignmentCost(dataSet[seqIndx], dataSet[q]);
        cost[q].push_back(seqML);
    }
        
    concepts.push_back(seqIndx);
    masks.push_back(true);    
    
    double ccptML = mmlEstimate(dataSet[seqIndx], mmlMatrix) + encodeWTNits(dataSet[seqIndx].size());
    
    nSize++;
    nSizeTrue++;
    dictML += ccptML;
    
    if(pertLegend){
        string pertLeg = "Add : " + to_string(seqIndx) + "\n";
        file_append(logF, pertLeg);
        cout << pertLeg;
    } 
    return true;  
}

void Dictionary::_remove(const int& indx){
    assert(indx > 0 && indx < nSize);
    assert(concepts.size() == nSize && masks.size() == nSize);
    assert(cost.size() == dataSet.size() && cost[0].size() == nSize);

    concepts.erase(concepts.begin() + indx);
    masks.erase(masks.begin() + indx);
    usages.erase(usages.begin() + indx);
    acceptedUsages.erase(acceptedUsages.begin() + indx);

    for(int i = 0; i < cost.size(); i++)
        cost[i].erase(cost[i].begin() + indx);
    
    nSize--; 
}

bool Dictionary::_swap(const int& indx){
    assert(indx > 0 && indx < nSize);
    assert(cost.size() == dataSet.size() && cost[0].size() == nSize);
    
    int usgSize = acceptedUsages[indx].size();
    if(usgSize == 1){
        _suppressConcept(indx);
        _remove(indx);
        return false;
    }
    
    int newind = 0;
    int r = rand() % (usgSize-1) + 1;
    assert(r > 0 && r < usgSize);
    newind = acceptedUsages[indx][r];
    assert(newind >= 0 && newind < dataSet.size());
    if(newind == concepts[indx] || dataSet[newind].compare(dataSet[concepts[indx]]) == 0) return false;

    tempCcpt = concepts[indx];
    concepts[indx] = newind;

    double seqML = 0;
    tempCost = vector<double>(dataSet.size(), 0);

    #pragma omp parallel for 
	for(size_t q = 0; q < dataSet.size(); q++){
        seqML = getAlignmentCost(dataSet[newind], dataSet[q]);
        tempCost[q] = cost[q][indx];
        cost[q][indx] = seqML;
    }      
	
    tempML = mmlEstimate(dataSet[tempCcpt], mmlMatrix) + encodeWTNits(dataSet[tempCcpt].size());
    dictML -= tempML;
    dictML += mmlEstimate(dataSet[newind], mmlMatrix) + encodeWTNits(dataSet[newind].size());
    
    if(pertLegend){
        string pertLeg = "Rep : " + to_string(tempCcpt) + "->" + to_string(newind) + "\n";
        file_append(logF, pertLeg);
        cout << pertLeg;
    }       
      
    return true;    
}

void Dictionary::_undoSwap(const int& indx){
    assert(indx > 0 && indx < nSize);
    assert(cost.size() == dataSet.size() && cost[0].size() == nSize);
    assert(tempCost.size() == dataSet.size());

    dictML -= mmlEstimate(dataSet[concepts[indx]], mmlMatrix) + encodeWTNits(dataSet[concepts[indx]].size());
    concepts[indx] = tempCcpt;
    dictML += tempML; 

    for(size_t q = 0; q < dataSet.size(); q++){
        cost[q][indx] = tempCost[q];
    }  
}

double Dictionary::getAlignmentCost(const string& ccpt, const string& sequence){
    int m = sequence.size();
    int n = ccpt.size();
    assert(m > 0 && n > 0);
    double inf = numeric_limits<double>::max();

    vector<vector<HistMatCell>> match_m, insert_m, delete_m;
    vector<vector<int>> pathCount;

    for(int k = 0; k < m+1; k++){
        vector<HistMatCell> temp1, temp2, temp3;
        for(int l = 0; l < n+1; l++){
            temp1.push_back(HistMatCell('M', k, l));
            temp2.push_back(HistMatCell('I', k, l));
            temp3.push_back(HistMatCell('D', k, l));
        }
        match_m.push_back(temp1);
        insert_m.push_back(temp2);
        delete_m.push_back(temp3);
    }

    match_m[0][0].cst = 0;
    insert_m[0][0].cst = 0;
    delete_m[0][0].cst = 0;

    match_m[0][0].backPtr = &match_m[0][0];
    insert_m[0][0].backPtr = &insert_m[0][0];
    delete_m[0][0].backPtr = &delete_m[0][0];


    match_m[1][0].cst = inf;
    insert_m[1][0].cst = mmlEstimateAA(sequence[0], mmlMatrix);
    delete_m[1][0].cst = inf;

    match_m[1][0].backPtr = &match_m[1][0];
    insert_m[1][0].backPtr = &insert_m[0][0];
    delete_m[1][0].backPtr = &delete_m[1][0];


    for(int i=2; i<=m; i++){
        match_m[i][0].cst = inf;
        insert_m[i][0].cst = i_i + mmlEstimateAA(sequence[i-1], mmlMatrix) + insert_m[i-1][0].cst;
        delete_m[i][0].cst = inf;

        match_m[i][0].backPtr = &match_m[i][0];
        insert_m[i][0].backPtr = &insert_m[i-1][0];
        delete_m[i][0].backPtr = &delete_m[i][0];
    }

    for(int j=1; j<=n; j++){
        match_m[0][j].cst = inf;
        insert_m[0][j].cst = inf;
        delete_m[0][j].cst = d_d*(j-1);

        match_m[0][j].backPtr = &match_m[0][j];
        insert_m[0][j].backPtr = &insert_m[0][j];
        delete_m[0][j].backPtr = &delete_m[0][j-1];
    }

    double scoreA, scoreB, scoreC;
    for(int i=1; i<=m; i++){
        for(int j=1; j<=n; j++){
            double sigma = conditEstimateAA(sequence[i-1], ccpt[j-1], condMatrix);
            double nVal = mmlEstimateAA(sequence[i-1], mmlMatrix);

            scoreA = match_m[i-1][j-1].cst + m_m + sigma;
            scoreB = insert_m[i-1][j-1].cst + m_i + sigma;
            scoreC = delete_m[i-1][j-1].cst + m_d + sigma;
            if(i == 1 && j == 1){
                scoreA -= m_m;
                scoreB -= m_i;
                scoreC -= m_d;
            }
            match_m[i][j].cst = min({scoreA, scoreB, scoreC});
            
            match_m[i][j].backPtr = (match_m[i][j].cst == scoreC) ? &delete_m[i-1][j-1] :
                                    (match_m[i][j].cst == scoreB) ? &insert_m[i-1][j-1] : 
                                    (match_m[i][j].cst == scoreA) ? &match_m[i-1][j-1] : &match_m[i][j];

            scoreA = match_m[i][j-1].cst  + d_m;
            scoreB = insert_m[i][j-1].cst + d_i;
            scoreC = delete_m[i][j-1].cst + d_d;
            delete_m[i][j].cst = min({scoreA, scoreB, scoreC});

            delete_m[i][j].backPtr = (delete_m[i][j].cst == scoreC) ? &delete_m[i][j-1] :
                                     (delete_m[i][j].cst == scoreB) ? &insert_m[i][j-1] : 
                                     (delete_m[i][j].cst == scoreA) ? &match_m[i][j-1] : &delete_m[i][j];

            scoreA = match_m[i-1][j].cst + i_m + nVal;
            scoreB = insert_m[i-1][j].cst + i_i + nVal;
            scoreC = delete_m[i-1][j].cst + i_d + nVal;
            insert_m[i][j].cst = min({scoreA, scoreB, scoreC}); 

            insert_m[i][j].backPtr = (insert_m[i][j].cst == scoreC) ? &delete_m[i-1][j] :
                                     (insert_m[i][j].cst == scoreB) ? &insert_m[i-1][j] : 
                                     (insert_m[i][j].cst == scoreA) ? &match_m[i-1][j] : &insert_m[i][j];
                                    
        }   
    }

    assert(match_m[m][n].cst > 0 && insert_m[m][n].cst > 0 && delete_m[m][n].cst > 0);
    int stateStringLen = printAlignment(ccpt, sequence, match_m, insert_m, delete_m);

    return min({match_m[m][n].cst,
                insert_m[m][n].cst,
                delete_m[m][n].cst
            }) - log(1/3.0) + encodeWTNits(stateStringLen - n); 
}

string Dictionary::traceback(string stateString, const HistMatCell* currCell) {
	if (currCell->i == 0 && currCell->j == 0) // if matrix (0,0) reached
		return stateString;
        
    stateString = currCell->type + stateString; 
	return traceback(stateString, currCell->backPtr); 
}

int Dictionary::printAlignment(const string& ccpt, const string& sequence, const vector<vector<HistMatCell>>& match_m, const vector<vector<HistMatCell>>& insert_m, const vector<vector<HistMatCell>>& delete_m){
    int i = sequence.size();
    int j = ccpt.size();
    
    std::vector<char> als;
    std::vector<char> alt;
    int stateStringLen = 0;
    double match, insertion, deletion;
    double msglen = 0;
    string stateStr = "";

    double min_score = min({match_m[i][j].cst, insert_m[i][j].cst, delete_m[i][j].cst});

    if(min_score == match_m[i][j].cst)
        stateStr = traceback("", &match_m[i][j]);
    else if (min_score == insert_m[i][j].cst)
		stateStr = traceback("", &insert_m[i][j]);
    else 
		stateStr = traceback("", &delete_m[i][j]);
        
    if(alignPrint) file_append(logF, "3-state String : " + stateStr); 

    i = 0, j = 0;
    vector<string> A; 
    for (size_t l = 0; l < stateStr.length(); l++) {
        if (stateStr[l] == 'I') {
            als.push_back(sequence[i]);
            alt.push_back('-');
            i++;
        } 
        else if (stateStr[l] == 'D') {
            als.push_back('-');
            alt.push_back(ccpt[j]);
            j++;
        }
        else if (stateStr[l] == 'M') {
            als.push_back(sequence[i]);
            alt.push_back(ccpt[j]);
            i++;
            j++;
        }
        else {
            cout << stateStr[l] << " " << "Something wierd!" << endl;
            assert(1);
        }
    }

    string ss = "", st = "", sa = "";
    vector<string> ssv, stv, sav;
    
    for(int i = 0; i < als.size(); i++){
        if(i != 0 && i%60 == 0) {
            ssv.push_back(ss);
            stv.push_back(st);
            sav.push_back(sa);
            ss = "", st = "", sa = "";
        }

        ss += als[i];
        st += alt[i];
        if(als[i] == alt[i]) sa += "*";
        else sa += " ";
    }
    
    ssv.push_back(ss);
    stv.push_back(st);
    sav.push_back(sa);
    
    string alignLog = "\n\nSequence  : " + sequence + "\n" + 
                      "Concept   : " + ccpt + "\n\n"
                      "Alignment : \n";

    assert(ssv.size() == stv.size() && ssv.size() == sav.size());
    for(int j = 0; j < ssv.size(); j++){
        alignLog += stv[j] + "\n" +  ssv[j] + "\n" + sav[j] + "\n\n";

    }
    if(alignPrint) file_append(logF, alignLog);

    return stateStr.size();
}


void Dictionary::_getML(double& dictmlF, double& datamlF){
    assert(nSize >= 1);
    if(nSize == 1) {
        dictmlF = bits2nits(1);
        assert(nullML > 0);
        datamlF = nullML;
    } else {
        dictmlF = dictML + encodeWTNits(nSizeTrue);
        datamlF = encodeWTNits(dataSet.size());

        usages.erase(usages.begin(), usages.end());
        for(int i = 0; i < nSize; i++){
            usages.push_back(vector<int>(1, -1));
        }
        
        for(size_t k = 0; k < dataSet.size(); k++){   //for each sequence in D: 
            double minCost = cost[k][0];
            int minCcpt = 0;
            for(int j = 1; j < cost[k].size(); j++){
                if(!masks[j]) continue;
                if(cost[k][j] <= minCost) {
                    minCost = cost[k][j];
                    minCcpt = j;
                }
            }

            if(seqCost) {
                string seqCostVal = "Sequence : " + to_string(k) + " - ( Concept " + to_string(minCcpt) + " : " + num2str(nits2bits(minCost), 8, 10) + ")\n";
                file_append(logF, seqCostVal);
                cout << seqCostVal;
            }       

            if(minCcpt > 0) usages[minCcpt].push_back(k);  
            datamlF += minCost + log(nSizeTrue); 
        }
    }
}

void Dictionary::_acceptUsages(){
    acceptedUsages = usages;
}


int Dictionary::_size(){
    return nSize-1;
}

void Dictionary::_checkPoint(const string& f_name){
    string content = "";

    for(const int& ccpt : concepts){
        if(ccpt < 0) continue;
        content += to_string(ccpt) + "\n";
    }
    file_append(f_name, content);
}

void Dictionary::_loadCheckPoint(const vector<string>& content){
    assert(content.size() > 1);
    int ind = -1;
    
    for(int j = 1; j < content.size(); j++){
        cout << j << endl;
        istringstream iss(content[j]);
        vector<string> legSplit((istream_iterator<string>(iss)),
                                 istream_iterator<string>());
        assert(legSplit.size() > 0);
        cout << setw(3) << j << " of " << content.size()-1 << " - ";
        ind = stoi(legSplit[0]);
        _add(ind);
    }
}

void Dictionary::_printUsage(vector<string>& content, const bool& wrt, const bool& dbrkdwn){
    assert(concepts.size() == acceptedUsages.size());

    string header1 = "\n----------------------- Dictionary Breakdown ------------------------";
    if(wrt) content.push_back(header1);
    if(dbrkdwn) cout << header1 << endl;

    for(int i = 0; i < acceptedUsages.size(); i++){
        if(acceptedUsages[i].size() <= 1) continue;
            
        stringstream is, usgs;
        is << setw(3) << concepts[i];
        usgs << setw(3) << acceptedUsages[i].size()-1;
        
        string ccptDesc = "# [" + is.str() + "] " + usgs.str() + " " + dataSet[concepts[i]];
        if(wrt) content.push_back(ccptDesc);
        if(dbrkdwn) cout << ccptDesc << endl;
        
    } 

    string footer = "---------------------------------------------------------------------\n";
    if(wrt) content.push_back(footer);
    if(dbrkdwn) cout << footer << endl << endl;

    string header3 = "\n------------------------- Usage Breakdown -------------------------";
    if(wrt) content.push_back(header3);
    if(dbrkdwn) cout << header3 << endl;

    for(int i = 0; i < acceptedUsages.size(); i++){
        if(acceptedUsages[i].size() <= 1) continue;

        stringstream indxs, is;
        indxs << setw(3) << (i+1);
        is << setw(3) << concepts[i];
        string temps = "# " + indxs.str() + " [" + is.str() + "]" + dataSet[concepts[i]] + "\n";

        for(int j = 1; j < acceptedUsages[i].size(); j++){
            stringstream ius;
            ius << setw(3) << acceptedUsages[i][j];
            temps += "\t@ [" + ius.str() + "] " + dataSet[acceptedUsages[i][j]] + "\n";
        } 
        if(dbrkdwn) cout << temps << endl;
        if(wrt) content.push_back(temps); 
   }
}
